from yta_general_utils.file.writer import FileWriter
from yta_general_utils.file.reader import FileReader
from yta_general_utils.programming.path import get_project_abspath

import json


CONFIG_MANIM_ABSPATH = get_project_abspath() + 'manim_parameters.json'

class ManimConfig:
    """
    Class to simplify and encapsulate the functionality
    related to manim configuration file, that is a file
    in which we write some configuration to be able to
    be read by the manim engine.
    """
    @classmethod
    def write(cls, json_data):
        """
        Writes the the provided 'json_data' manim configuration
        in the configuration file so the parameters could be read
        later by the manim engine. This is the way to share 
        parameters to the process.
        """
        # TODO: Check that 'json_data' is valid and well-formatted
        FileWriter.write_file(json.dumps(json_data, indent = 4), CONFIG_MANIM_ABSPATH)

    @classmethod
    def read(cls):
        """
        Read the configuration file and return it as a json
        object.
        """
        return FileReader.read_json_from_file(CONFIG_MANIM_ABSPATH)


from yta_multimedia.alphascreen.masked_clip_creator import MaskedClipCreator
from yta_multimedia.image.region.image_region import ImageRegion
from yta_multimedia.image.region.image_region_finder import ImageRegionFinder
from yta_general_utils.image.checker import has_transparency
from yta_general_utils.file.checker import is_valid_image
from moviepy import ImageClip
from PIL import Image


class ImageAlphascreen:
    """
    Class to handle images with alphascreen regions and insert
    other images or videos on it.
    """
    image = None
    image_filename: str = None
    alpha_regions: list[ImageRegion] = []

    def __init__(self, filename: str):
        if not isinstance(filename, str):
            raise Exception(f'No str "filename" parameter "{filename}" provided.')
        
        if not is_valid_image(filename):
            raise Exception(f'The provided "filename" parameter "{filename}" is not a valid image.')
        
        image = Image.open(filename)

        if not has_transparency(image):
            raise Exception('The provided image "filename" parameter "{filename}" does not have any alpha channel.')

        self.image_filename = filename
        self.image = image
        self.alpha_regions = ImageRegionFinder.find_transparent_regions(self.image)

        if len(self.alpha_regions) == 0:
            raise Exception('No alpha regions found in the "filename" parameter "{filename}" provided.')
        
        # TODO: What about regions that are just one pixel or too short (?)

        # Duration will be processed and updated in the last step
        alpha_clip = ImageClip(self.image_filename, duration = 1 / 60)
        self.masked_clip_creator = MaskedClipCreator(self.alpha_regions, alpha_clip)
        
    def from_image_to_image(self, image, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'image' in this alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        # TODO: This is not returning RGBA only RGB
        return self.masked_clip_creator.from_image_to_image(image, output_filename)
    
    def from_images_to_image(self, images, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'images' in this
        alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_images_to_image(images, output_filename)
    
    def from_image_to_video(self, image, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'image' fitting the first alphascreen area and centered on
        those areas by applying a mask that let them be seen
        through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_images_to_video(image, duration, output_filename)

    def from_images_to_video(self, images, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'images' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_images_to_video(images, duration, output_filename)
    
    def from_video_to_video(self, video, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'video' fitting in the alphascreen area and centered on it
        by applying a mask that let it be seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_video_to_video(video, output_filename)
    
    def from_videos_to_video(self, videos, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'videos' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_videos_to_video(videos, output_filename)


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math import Math
from moviepy import Clip, clips_array

import math


class MultipliedEffect(Effect):
    """
    Generates a clips array with the provided 'video' being shown
    'times' times (this parameter must be a pow of 4). This
    method has been created to be used internally with our own
    default methods.
    """
    def apply(self, video: Clip, times: int) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if not times:
            times = 4
        else:
            if not Math.is_power_of_n(times, 4):
                raise Exception(f'The provided "times" parameter "{str(times)}" is not a power of 4.')

        audio = video.audio
        size = (video.w, video.h)

        # We build the matrix of sqrt(times, 2) videos
        # per row, and sqrt(times, 2) rows. So, if 16
        # 'times' provided, that is power of 4 and valid,
        # it will generate a 4x4 array. If 64 'times',
        # a 8x8 array that will be the composition
        items = int(math.sqrt(times))
        array = [[video] * items for _ in range(items)]

        return clips_array(array).resized(size).with_audio(audio)


from yta_general_utils.temp import create_temp_filename
from yta_general_utils.downloader import Downloader
from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from PIL import Image, ImageFont, ImageDraw, ImageOps
from random import randrange
from typing import Union

import textwrap
import numpy as np


# TODO: Maybe implement it as a Singleton like DiscordImageGenerator
# to be faster when reusing the scrapper for consecutive generations
class TripadvisorImageGenerator:
    def generate_review_manually(self, review: str = None, avatar_url: str = None, username: str = None, city: str = None, contributions_number: int = randrange(30), rating_stars: int = (randrange(5) + 1), title: str = None, output_filename: str = None):
        # TODO: This must be set by the user in .env I think
        __FONTS_PATH = 'C:/USERS/DANIA/APPDATA/LOCAL/MICROSOFT/WINDOWS/FONTS/'

        # TODO: Do more checkings please
        # TODO: Check contributions_number is in between 1 and 30 and is a number
        # TODO: Check rating_stars is a number between 1 and 5
        # TODO: Check title exists
        # TODO: Check review exists

        bold_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST_777_BOLD_BT.TTF', 22, encoding = 'unic')
        normal_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST-777-FONT.OTF', 16, encoding = 'unic')
        review_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST-777-FONT.OTF', 20, encoding = 'unic')

        # Create white review background (440 is h, 800 is w)
        img = np.zeros((440, 800, 3), dtype = np.uint8)
        img.fill(255)
        img = Image.fromarray(img).convert('RGBA')
        draw = ImageDraw.Draw(img)

        # Place avatar image circle at (24, 31)
        image_filename = create_temp_filename('tmp_avatar.png')
        if not avatar_url:
            avatar_url = 'https://avatar.iran.liara.run/public'
        Downloader.download_image(avatar_url, image_filename)
        avatar_image = Image.open(image_filename)

        # Make image fit and then mask with circle
        avatar_image = ImageOps.fit(avatar_image, (64, 64))
        mask = Image.new('L', avatar_image.size, 0)
        circle_draw = ImageDraw.Draw(mask)
        circle_draw.ellipse((0, 0, 64, 64), fill = 255)
        result = Image.new('RGBA', avatar_image.size, (255, 255, 255))
        result.paste(avatar_image, (0, 0), mask)

        img.paste(result, (24, 31))

        # I check if I need some fake information
        fake_data = None
        if not username or not city:
            # TODO: Move this below to a 'get_json_from_url' or similar
            # TODO: Create the new 'yta-data-provider' library and include this
            import requests
            response = requests.get('https://fakerapi.it/api/v1/persons?_locale=es_ES')
            data = response.json()
            fake_data = data['data'][0] # First person generated data

        # Put reviewer name at (103, 33)
        if not username:
            username = fake_data['firstname'] + ' ' + fake_data['lastname']
        draw.text((103, 33), username, font = bold_font, fill = (84, 84, 84), line_height = '19px', width = 700)
        
        # Write city and contributions at (103, 65)
        if not city:
            city = fake_data['address']['city']
        # TODO: Check contributions_number is a number
        draw.text((103, 65), city + ' · ' + str(contributions_number) + ' contribuciones', font = normal_font, fill = (84, 84, 84), line_height = '19px', width = 700)

        # Build the evaluation (in green circles) at (24, 126)
        def __draw_filled_ellipse(xy):
            return draw.ellipse(xy, fill = (0, 170, 108), outline = (0, 170, 108))
        def __draw_unfilled_ellipse(xy):
            return draw.ellipse(xy, fill = (255, 255, 255), outline = (0, 170, 108))
        
        x = 24
        for i in range(5):
            if i < rating_stars:
                __draw_filled_ellipse((x, 126, x + 24, 126 + 24))
            else:
                __draw_unfilled_ellipse((x, 126, x + 24, 126 + 24))
            x += 24 + 6

        # Place the review title at (24, 164)
        if not title:
            # TODO: Fake review title with AI
            title = 'Menuda pasada'
        draw.text((24, 164), title, font = bold_font, fill = (84, 84, 84), line_height = '19px')

        # Place the review date at (24, 199)
        # TODO: Implement 'date' parameter
        # TODO: Implement 'visit_type' options
        draw.text((24, 199), "jul 2019 · En pareja", font = normal_font, fill = (84, 84, 84), line_height = '19px')

        # Place the whole wrapped text starting at (24, 440 - 180)
        # See this: (https://gist.github.com/turicas/1455973) to wrap text
        if not review:
            # TODO: Fake review with AI
            review = 'Estuvimos todo el viaje pensando en que ibamos a poder visitar por fin este sitio, y la verdad es que no ha defraudado nada. Lo disfrutamos de principio a fin y la atención fue espectacular. Llevábamos unas espectativas bastante altas pero, sin lugar a dudas, las han superado. Encantados!!'
        lines = textwrap.wrap(review, width = 76)
        y_text = 440 - 180
        for line in lines:
            left, top, right, bottom = review_font.getbbox(line)
            height = bottom - top
            draw.text((24, y_text), line, font = review_font, fill = (84, 84, 84))
            y_text += height + 8

        # Place the number of likes: 700, 50
        #draw.text((700, 50), "3", font = normal_font, fill = (84, 84, 84), line_height = '19px')

        if not output_filename:
            output_filename = create_temp_filename('tmp_tripadvisor_review.png')
        img.save(output_filename, 'png')

        # TODO: Why not returning the image instead (?)
        return output_filename
    
    def generate_review(self, username: str = None, user_picture_url: str = None, title: str = None, message: str = None, rating: int = None, output_filename: Union[str, None] = None):
        if not username:
            # TODO: Fake it
            username = 'Gabriel de las Salinas'

        if not user_picture_url:
            # TODO: Fake it
            user_picture_url = 'https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcTmIVOqsYK3t8HxkQ_WjwPoP2cwJiV1xDyWIw&s'

        if not title:
            # TODO: Fake it
            title = 'Estancia un poco aleatoria'

        if not message:
            # TODO: Fake it
            message = 'Este es un mensaje puesto a dedo en un comentario, que quiero que sea largo para que ocupe dos líneas a ser posible, y así lo veamos bien cómo quedaría.'

        if not rating:
            rating = randrange(1, 5)

        # TODO: Implement date
        # TODO: Implement review type ('Pareja', 'Solitario', 'Familia', etc.)

        if rating < 1:
            rating = 1
        elif rating > 5:
            rating = 5

        # I don't let manually handle this value (it is useless)
        contributions_number = randrange(1, 15)
        # TODO: Implement this in the future (maybe)
        # likes_number = 13

        # TODO: I don't know why this is not working without GUI
        scrapper = ChromeScrapper(True)
        scrapper.go_to_web_and_wait_util_loaded('https://www.tripadvisor.es/AttractionProductReview-g189180-d25382579-The_Unvanquished_Tour_in_Porto_City_Center-Porto_Porto_District_Northern_Portugal.html')
        # Get <div data-automation="reviewCard">
        review_element = scrapper.find_element_by_custom_tag_waiting('div', 'data-automation', 'reviewCard')
        review_card_divs = scrapper.find_elements_by_element_type('div', review_element, only_first_level = True)
        scrapper.scroll_to_element(review_element)
        scrapper.scroll_up(100)

        # User image url   divx8 > a > picture > img.src = new_url
        image_element = review_card_divs[0]
        for i in range(7):
            image_element = scrapper.find_element_by_element_type('div', image_element)
        image_element = scrapper.find_element_by_element_type('a', image_element)
        image_element = scrapper.find_element_by_element_type('div', image_element)
        image_element = scrapper.find_element_by_element_type('picture', image_element)
        image_element = scrapper.find_element_by_element_type('img', image_element)
        scrapper.set_element_attribute(image_element, 'src', user_picture_url)

        # Username   divx2 > div#1 (2o div) > span > a inner_text
        author_element = review_card_divs[0]
        author_element = scrapper.find_element_by_element_type('div', author_element)
        author_element = scrapper.find_elements_by_element_type('div', author_element, only_first_level = True)[1]
        # TODO: This [1] is doing div > div, not getting the second 
        # child in horizontal hierarchy
        author_element = scrapper.find_element_by_element_type('span', author_element)
        author_element = scrapper.find_element_by_element_type('a', author_element)
        scrapper.set_element_inner_text(author_element, username)

        # Contributions   divx2 > div#1 (2o div) > divx2 > span inner_text
        contributions_element = review_card_divs[0]
        contributions_element = scrapper.find_element_by_element_type('div', contributions_element)
        contributions_element = scrapper.find_elements_by_element_type('div', contributions_element, only_first_level = True)[1]
        contributions_element = scrapper.find_element_by_element_type('div', contributions_element)
        contributions_element = scrapper.find_element_by_element_type('div', contributions_element)
        contributions_element = scrapper.find_element_by_element_type('span', contributions_element)
        # 1 contribución, X contribuciones
        contributions_str = '1 contribución'
        if contributions_number > 1:
            contributions_str = str(contributions_number) + ' contribuciones'
        scrapper.set_element_inner_text(contributions_element, contributions_str)

        # Rating (starts alike)   div#1 (2o div) > svg > path (special)
        rating_element = review_card_divs[1]
        rating_element = scrapper.find_element_by_element_type('svg', rating_element)
        paths = scrapper.find_elements_by_element_type('path', rating_element)
        FILLED_CIRCLE_D_VALUE = 'M 12 0C5.388 0 0 5.388 0 12s5.388 12 12 12 12-5.38 12-12c0-6.612-5.38-12-12-12z'
        UNFILLED_CIRCLE_D_VALUE = 'M 12 0C5.388 0 0 5.388 0 12s5.388 12 12 12 12-5.38 12-12c0-6.612-5.38-12-12-12zm0 2a9.983 9.983 0 019.995 10 10 10 0 01-10 10A10 10 0 012 12 10 10 0 0112 2z'
        for i in range(5):
            if rating > i:
                scrapper.set_element_attribute(paths[i], 'd', FILLED_CIRCLE_D_VALUE)
            else:
                scrapper.set_element_attribute(paths[i], 'd', UNFILLED_CIRCLE_D_VALUE)

        # Title   div#2 (3er div) > a > span inner_text
        title_element = review_card_divs[2]
        title_element = scrapper.find_element_by_element_type('a', title_element)
        title_element = scrapper.find_element_by_element_type('span', title_element)
        scrapper.set_element_inner_text(title_element, title)

        # Date and type of visit   div#3 (4o div) inner_text
        date_and_type_element = review_card_divs[3]
        # TODO: Work this below, please
        # Accepted values: Pareja, Familia
        # example: sept de 2024 • Familia
        date_and_type_str = 'sept de 2024 • Pareja'
        scrapper.set_element_inner_text(date_and_type_element, date_and_type_str)

        # Message (opinion)   div#4 (5o div) > div > div > span > span inner_text (with <br> please)
        message_element = review_card_divs[4]
        message_element = scrapper.find_element_by_element_type('div', message_element)
        message_element = scrapper.find_element_by_element_type('div', message_element)
        message_element = scrapper.find_element_by_element_type('span', message_element)
        message_element = scrapper.find_element_by_element_type('span', message_element)
        scrapper.set_element_inner_text(message_element, message)

        # Images is the div#4 (5o div) but it is empty if no pictures
        # and we don't care (by now) about this div, just to remove it
        scrapper.remove_element(review_card_divs[5])

        # Written date   div#5 (6o div) > div#0 inner_text
        written_date_element = review_card_divs[6]
        written_date_element = scrapper.find_element_by_element_type('div', written_date_element)
        # TODO: Handle this please
        written_date_str = 'Escrita el 2 de septiembre de 2024'
        scrapper.set_element_inner_text(written_date_element, written_date_str)

        # Tripadvisor disclaimer   div#5 (6o div) > div#1 inner_text
        tripadvisor_disclaimer_element = written_date_element = review_card_divs[len(review_card_divs) - 1]
        tripadvisor_disclaimer_element = scrapper.find_elements_by_element_type('div', written_date_element, only_first_level = True)[1]
        disclaimer_str = 'Esta es una opinión personal y no de la cuenta que publica este contenido.'
        scrapper.set_element_inner_text(tripadvisor_disclaimer_element, disclaimer_str)

        style = 'width: 500px; padding: 10px;'
        scrapper.set_element_style(review_element, style)

        if not output_filename:
            output_filename = create_temp_filename('tmp_youtube_thumbnail.png')

        scrapper.screenshot_element(review_element, output_filename)

        return Image.open(output_filename)



from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text
from manim import *


class TextTripletsManimAnimation(BaseManimAnimation):
    """
    The provided 'text' is splitted in triplets and appear on the screen. This animation
    lasts 'duration' seconds. Each triplet appear each 'duration' / len(words).
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, text: str, duration: float, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('text', self.required_parameters) and not text:
            raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
        parameters['text'] = text

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename

    def animate(self):
        words = self.parameters['text'].split(' ')
        # We need to adjust the array to contain a multiple of 3 number of elements
        leftover_numbers = words[len(words) - len(words) % 3:]
        if len(leftover_numbers) > 0:
            words = words[:len(words) - len(leftover_numbers)]

        words_triplets = []
        subarray = []
        for word in words:
            subarray.append(word)
            if len(subarray) == 3:
                words_triplets.append(subarray)
                subarray = []
        if leftover_numbers:
            words_triplets += [leftover_numbers]
        each_triplet_time = self.parameters['duration'] / len(words_triplets)

        for triplet in words_triplets:
            str = ' '.join(triplet)
            # I don't know how to show one word before the other
            # Helping information:
            # For example say you would like to not render and skip the begin of a video , you put self.next_section(skip_animations=True) in the line after def construct(self): and put self.next_section() before the first line of the animation you want to render.
            # Thank you: https://www.reddit.com/r/manim/comments/tq1ii8/comment/ihzogki/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
            text = fitting_text(str, 1920 / 2)
            # Create 3 similar texts with each word
            self.add(text)
            self.wait(each_triplet_time)
            self.remove(text)


    


import pyttsx3


def narrate(text, output_filename = None):
    """
    Creates an audio narration of the provided 'text' and stores it as 'output_filename'.
    """
    if not output_filename:
        return None
    
    # TODO: This is hardcoded, be careful!
    SPANISH_SPAIN_VOICE_ID = 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Voices\Tokens\TTS_MS_ES-ES_HELENA_11.0'
    SPANISH_MEXICO_VOICE_ID = 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Voices\Tokens\TTS_MS_ES-MX_SABINA_11.0'
    engine = pyttsx3.init()
    engine = pyttsx3.init()
    engine.setProperty('voice', SPANISH_SPAIN_VOICE_ID)
    # Default speed is 200 (wpm)
    engine.setProperty('rate', 130)
    engine.save_to_file(text, output_filename)
    engine.runAndWait()


"""
This module is to encapsulate some factor calculations
to make easy building new effects.
"""
from typing import Union
from yta_multimedia.video.edition.effect.moviepy.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_multimedia.image.region.image_region import ImageRegion


def get_factor_to_fit_scene(video_position: Union[tuple, Coordinate, Position], position_in_scene: Union[tuple, Coordinate, Position], video_size: tuple, scene_size: tuple = (1920, 1080)):
    """
    This method calculates the factor we need to apply to the
    resize method to obtain a video that fits the scene without
    any black region.
    """
    if not PythonValidator.is_instance(video_position, [Position, Coordinate]):
        if not PythonValidator.is_instance(video_position, tuple) and len(video_position) != 2:
            raise Exception('Provided "video_position" is not a valid Position enum or (x, y) tuple.')
        else:
            video_position = Coordinate(video_position[0], video_position[1])
        
    if not PythonValidator.is_instance(position_in_scene, [Position, Coordinate]):
        if not PythonValidator.is_instance(position_in_scene, tuple) and len(position_in_scene) != 2:
            raise Exception('Provided "position_in_scene" is not a valid Position enum or (x, y) tuple.')
        else:
            position_in_scene = Coordinate(position_in_scene[0], position_in_scene[1])

    # TODO: Improve this by checking that is about positive values
    if not PythonValidator.is_tuple(video_size) or len(video_size) != 2:
        raise Exception('The provided "video_size" is not a valid size.')
    
    if not PythonValidator.is_tuple(scene_size) or len(scene_size) != 2:
        raise Exception('The provided "scene_size" is not a valid size.')
    
    # We need to calculate the difference in size between the 
    # scene size and the video size with the new position as
    # its center to be able to calculate the factor later
    video_and_scene_size_difference = (
        scene_size[0] / (video_size[0] - abs(position_in_scene.get_moviepy_center_tuple()[0] - video_position.x) * 2),
        scene_size[1] / (video_size[1] - abs(position_in_scene.get_moviepy_center_tuple()[1] - video_position.y) * 2)
    )

    # We calculate the factor we need to apply to resize the
    # video to fit the actual scene size and avoid any black
    # region
    factor = max(
        scene_size[0] / video_and_scene_size_difference[0],
        scene_size[1] / video_and_scene_size_difference[1]
    )

    return factor

def get_factor_to_fit_region(region: ImageRegion, scene_size: tuple = (1920, 1080)):
    """
    This method calculates the factor we need to apply to the
    resize method to obtain a video in which the provided 
    'region' is displayed fitting the whole scene with its
    center in the middle of it.
    """
    if not PythonValidator.is_instance(region, ImageRegion):
        raise Exception('The provided "region" is not a valid Region class instance.')
    
    if not PythonValidator.is_tuple(scene_size) or len(scene_size) != 2:
        raise Exception('The provided "scene_size" is not a valid size.')
    
    # We calculate the resize factor we need to make the
    # video fit only the region size
    factor = max(
        scene_size[0] / region.width,
        scene_size[1] / region.height
    )

    return factor


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyArgument, MoviepyWithPrecalculated
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionResize
from yta_multimedia.video import VideoHandler
from yta_multimedia.video.edition.effect.moviepy.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.edition.effect.moviepy.position.utils.factor import get_factor_to_fit_scene
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions import RateFunction
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.Clip import Clip
from typing import Union

import numpy as np

# TODO: Rename it, as it is making zoom but in an
# specific an static position, so it should be
# something like ZoomInPlaceEffect
class ZoomOnVideoPositionEffect(Effect):
    """
    Creates a Zoom using a specific position of the video
    as its center to make the zoom effect on it. This 
    effect will adapt the video size to fit the full scene
    size and avoid showing any black area.
    """
    def apply(cls, video: Clip, video_position: Union[tuple, Coordinate, Position], position_in_scene: Union[tuple, Coordinate, Position], rate_func: type = RateFunction.linear):
        """
        Apply the effect on the provided 'video'. The 'video_position'
        parameter is the position (within the video) we want to use as
        the new center. The 'position_in_scene' is the position (within
        the scene) in which we want to place the new center of the
        provided 'video'.
        """
        # TODO: What about 'rate_func' (?)
        video = VideoParser.to_moviepy(video)

        if not PythonValidator.is_instance(video_position, [Position, Coordinate]):
            if not PythonValidator.is_instance(video_position, tuple) and len(video_position) != 2:
                raise Exception('Provided "video_position" is not a valid Position enum or (x, y) tuple.')
            else:
                video_position = Coordinate(video_position[0], video_position[1])
            
        if not PythonValidator.is_instance(position_in_scene, [Position, Coordinate]):
            if not PythonValidator.is_instance(position_in_scene, tuple) and len(position_in_scene) != 2:
                raise Exception('Provided "position_in_scene" is not a valid Position enum or (x, y) tuple.')
            else:
                position_in_scene = Coordinate(position_in_scene[0], position_in_scene[1])

        # TODO: This is actually written in another part of the code
        # and maybe I can make it dynamic depending on the actual
        # background (scene) size, but by now here it is to make it 
        # work
        MAX_SCENE_SIZE = (1920, 1080)

        resize_factor = get_factor_to_fit_scene(video_position, position_in_scene, (video.w, video.h), MAX_SCENE_SIZE)

        video_handler = VideoHandler(video)
        #gsvideo = gsvideo.resized(min_resize_factor)
        # TODO: This 1.5 is the static zoom factor we are using
        # now but it is only for this zoom. Make it dynamic later
        resizes = [
            TFunctionResize.resize_from_to(t, video.duration, resize_factor * 1.5, resize_factor, RateFunction.linear) for t in video_handler.frame_time_moments
        ]

        # Calculate new differente between the real center 
        # and the new center of the video
        video_center_difference = (
            video.w - video_position.x * 2,
            video.h - video_position.y * 2
        )

        positions = [
            position_in_scene.get_moviepy_upper_left_corner_tuple((
                (video.w - video_center_difference[0]) * resizes[i],
                (video.h - video_center_difference[1]) * resizes[i]
            )) for i in range(len(resizes))
        ]

        return MoviepyWithPrecalculated().apply(video, with_position_list = positions, resized_list = resizes)

    # TODO: 'apply_over_video' here (?)


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from moviepy.video.fx import MirrorX
from moviepy import Clip


class FlipHorizontallyEffect(Effect):
    """
    This effect flips the video horizontally (including
    the mask).
    """
    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        return MirrorX().apply(video)


from yta_general_utils.file.filename import get_file_filename_without_extension
from subprocess import run


class AudioSeparator:
    """
    Class to simplify and encapsulate the functionality related to
    extracting vocals, instruments and more from a single audio.
    """
    def separate_audio_file_into_vocals_and_instruments(audio_filename: str, output_dirname: str):
        """
        This method will get 'audio_filename' audio file and will split
        it into 2 different files stored in 'output_dirname' directory.
        These 2 files will be 'vocals.wav' and 'accompaniment.wav' that
        are the vocals and the instrumental part, respectively, and 
        will be generated in a folder inside the 'output_dirname' 
        provided that is the 'audio_filename' file name without the
        extension, and will be returned by the function.

        If 'example.wav' provided as 'audio_filename' and 'test' as 
        'output_dirname', this method will create 
        'test/example/vocals.wav' and 'test/example/accompaniment.wav'
        audio files.
        """
        # This is using Spleeter, check this (https://github.com/deezer/spleeter/wiki/2.-Getting-started)
        #
        # This method needs a pre-trained model (that is downloaded from the repo)
        # for each type (2stems, 4stems or 5stems). These are downloaded by default
        # in 'pretrained_models' folder (project root) of the project in which code
        # is being executed
        #
        # You can also do 'spleeter:4stems' and 'spleeter:5stems'
        # 2stems ->   vocals.wav, accompaniment.wav
        # 4stems ->   vocals.wav, drums.wav, bass.wav, piano.wav
        # 5stems ->   vocals.wav, drums.wav, bass.wav, piano.wav and other.wav
        #
        # By now this method is only working for the '2stems' format so it will
        # generate 'vocals.wav' and 'accompaniment.wav' in the provided 
        # 'output_dirname'. We could personalize this in the future.
        
        # TODO: We should accept any type of audio and, if needed, we
        # write it to a local file to be read by the spleeter command
        if not audio_filename:
            return None
        
        # TODO: Make this dynamic (by using temporary values)
        if not output_dirname:
            return None
        
        command_str = 'spleeter separate -p spleeter:2stems -o ' + output_dirname + ' ' + audio_filename

        try:
            run(command_str)
        except:
            return False

        if not output_dirname.endswith('/'):
            output_dirname += '/'

        generated_output_folder = output_dirname + get_file_filename_without_extension(audio_filename)
        
        return generated_output_folder

        # TODO: Please, make it work with python code, this below is not working
        """
        from spleeter.audio.adapter import AudioAdapter
        from spleeter.separator import Separator

        separator = Separator('spleeter:2stems')
        # With this below we read file as numpy array (waveform parameter)
        #from scipy.io.wavfile import read
        #audiowave = read(audio_filename)
        audio_adapter = AudioAdapter.default()
        separator.separate_to_file(audio_descriptor = audio_filename, destination = output_dirname, audio_adapter = audio_adapter)
        separator.join()
        """


from yta_multimedia.video.generation.manim.constants import SCENE_HEIGHT, SCENE_WIDTH, STANDARD_HEIGHT, STANDARD_WIDTH
from manim import *


class ManimDimensions:
    """
    Class to encapsulate and simplify the functionality related to manim
    dimensions. Manim works with specific dimensions that are not measured
    in pixels and we use to work with pixels. This class was created to
    let us work with pixel dimensions that will be translated into manim
    dimensions to simplify the creative process.

    We consider that a manim screen has 1920x1080 dimensions in pixels, so
    those are the pixel dimensions you should keep in mind for your
    calculations.

    Manim dimensions system is (14 + 2/9)w x (8)h, which means that the 
    width is 14 + 2/9 and the height is 8. The upper left corner coordinate
    is [-7-1/9, 4, 0] (0 is the z axis), and the lower right corner 
    coordinate is [7+1/9, -4, 0], because the center of the scene and screen
    is the origin [0, 0, 0].
    """
    @staticmethod
    def width_to_manim_width(width: float):
        """
        Turns the pixel 'width' provided dimension to the corresponding manim
        width dimension. Remember that the system is prepared to work with a
        simulated screen of 1920x1080 pixels, so providing a 'width' of 1920
        will make the object fit the whole manim screen width.
        """
        return (width * SCENE_WIDTH) / STANDARD_WIDTH
    
    @staticmethod
    def manim_width_to_width(manim_width: float):
        """
        Turns the provided 'manim_width' dimension to the corresponding pixel
        width dimension based on our simulated screen size of 1920x1080 pixels.
        That means that providing a 'manim_width' of 14 + 2/9 will return 1920
        as pixel width, that is the screen size for our simulated whole screen.
        """
        # TODO: Maybe check if 'manim_width' is too big to raise
        # an exception or, at least, print a message
        return (manim_width * STANDARD_WIDTH) / SCENE_WIDTH
    
    @staticmethod
    def height_to_manim_height(height: float):
        """
        Turns the pixel 'height' provided dimension to the corresponding manim
        height dimension. Remember that the system is prepared to work with a
        simulated screen of 1920x1080 pixels, so providing a 'height' of 1080
        will make the object fit the whole manim screen height.
        """
        return (height * SCENE_HEIGHT) / STANDARD_HEIGHT
    
    @staticmethod
    def manim_height_to_height(manim_height: float):
        """
        Turns the provided 'manim_height' dimension to the corresponding pixel
        height dimension based on our simulated screen size of 1920x1080 pixels.
        That means that providing a 'manim_height' of 8 will return 1080 as
        pixel height, that is the screen size for our simulated whole screen.
        """
        # TODO: Maybe check if 'manim_height' is too big to raise
        # an exception or, at least, print a message
        return (manim_height * STANDARD_HEIGHT) / SCENE_HEIGHT



# TODO: Rename this
class ManimXGenerator:
    """
    Class to simplify and encapsulate the mobject generation functionality.
    """
    @staticmethod
    def mobject_fitting_width(mobject: Mobject, width: float, **kwargs):
        """
        Creates the provided 'mobject' with a size that fits the provided
        'width'. This will be limited to the height. If the new mobject
        height is greater than the screen max height, it will be limited
        to the height.

        If you want a mobject that just fits the provided 'width' ignoring
        the height, just use 'mobject.scale_to_fit_width(width)' instead.
        """
        width = ManimDimensions.width_to_manim_width(width)

        # We build both width and height fitted and get the most limited
        mobject_width_fitted = mobject(**kwargs).scale_to_fit_width(width)
        mobject_height_fitted = mobject(**kwargs).scale_to_fit_height(width)

        # As it is a 16:9 proportion, the height is the measure that limits the most
        if mobject_height_fitted.width < mobject_width_fitted.width:
            return mobject_height_fitted
        
        return mobject_width_fitted
    
    def mobject_fitting_height(mobject: Mobject, height: float, **kwargs):
        """
        Creates the provided 'mobject' with a size that fits the provided
        'height'. This will be limited to the width. If the new mobject
        width is greater than the screen max width, it will be limited
        to the width.

        If you want a mobject that just fits the provided 'height' ignoring
        the width, just use 'mobject.scale_to_fit_height(height)' instead.
        """
        height = ManimDimensions.height_to_manim_height(height)

        # We build both width and height fitted and get the most limited
        mobject_height_fitted = mobject(**kwargs).scale_to_fit_height(height)
        mobject_width_fitted = mobject(**kwargs).scale_to_fit_width(height)

        # As it is a 16:9 proportion, the height is the measure that limits the most
        # TODO: When it was Text the mobject I used the font.size to compare
        if mobject_height_fitted.width < mobject_width_fitted.width:
            return mobject_height_fitted
        
        return mobject_width_fitted
    
    def mobject_fitting_fullscreen(mobject: Mobject, **kwargs):
        """
        Creates the provided 'mobject' with a size to fit the whole screen.
        This method will return the provided 'mobject' fitting the standard
        whole screen. The mobject will be cropped if its aspect ratio is not 
        16:9, but it will fit the whole screen for sure.
        """
        image_width_fitted = mobject(**kwargs).scale_to_fit_width(ManimDimensions.width_to_manim_width(STANDARD_WIDTH))

        # We want the mobject that occupies the whole screen
        if ManimDimensions.manim_height_to_height(image_width_fitted.height) >= STANDARD_HEIGHT:
            return image_width_fitted
        
        image_height_fitted = mobject(**kwargs).scale_to_fit_height(ManimDimensions.height_to_manim_height(STANDARD_HEIGHT))

        return image_height_fitted
    
    def mobject_fitting_screen(mobject: Mobject, **kwargs):
        """
        Scales the provided 'mobject' if necessary to fit the screen. That 
        means that a mobject bigger than the screen size will be cropped
        to fit inside. One of the two dimensions (width or height) could
        be out of bounds if the provided 'mobject' is bigger than the 
        screen size when provided.
        """
        mobject = mobject(**kwargs)

        if ManimDimensions.manim_width_to_width(mobject.width) > STANDARD_WIDTH:
            mobject.scale_to_fit_width(ManimDimensions.width_to_manim_width(STANDARD_WIDTH))
        if ManimDimensions.manim_height_to_height(mobject.height) > STANDARD_HEIGHT:
            mobject.scale_to_fit_height(ManimDimensions.height_to_manim_height(STANDARD_HEIGHT))

    # TODO: Maybe we don't have to instantiate the mobject but resizing
    # it so 'mobject(**kwargs)' is not the way.


# TODO: You can do this below so you don't need to declare all args
# just pass the **kwargs and let the user build the 'color', 
# 'font_size' and all params he wants.
"""
kwargs = {
    'color': BLUE,
    'font_size': 22
}
self.play(Text('hola', **kwargs).animate.rotate(90 * DEGREES), run_time = 2)
"""

# TODO: Remove this below when refactored and unneeded
def fitting_text(text, width_to_fit: float = 1920, fill_opacity: float = 1, stroke_width: float = 0, color: ParsableManimColor = None, font_size: float = DEFAULT_FONT_SIZE, line_spacing: float = -1, font: str = '', slant: str = NORMAL, weight: str = NORMAL, t2c: dict[str, str] = None, t2f: dict[str, str] = None, t2g: dict[str, tuple] = None, t2s: dict[str, str] = None, t2w: dict[str, str] = None, gradient: tuple = None, tab_width: int = 4, warn_missing_font: bool = True, height: float = None, width: float = None, should_center: bool = True, disable_ligatures: bool = False, **kwargs):
    """
    This method returns a Text mobject that fits the provided 'width_to_fit'
    or, if the height is greater than the scene height, returns one with the
    greates possible width.
    
    This method has been built to be sure that your text is completely shown
    between the screen margins.

    @param
        **width_to_fit**
        The widht you want to fit, in normal pixels (1920 is the maximum). 
        These pixels will be processed to manim dimensions.
    """
    width_to_fit = ManimDimensions.width_to_manim_width(width_to_fit)

    txt_width_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_width(width_to_fit)
    # I use a margin of 100 pixels so avoid being just in the borders
    txt_height_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_height(SCENE_HEIGHT - ManimDimensions.height_to_manim_height(100))

    # As it is a 16:9 proportion, the height is the measure that limits the most
    if txt_height_fitted.font_size < txt_width_fitted.font_size:
        return txt_height_fitted
    return txt_width_fitted

def fitting_image(filename, width_to_fit, image_mode: str = 'RGBA', **kwargs):
    """
    Returns an ImageMobject of the provided 'filename' image that fits the provided 'width_to_fit'
    or, if the height limit is surpassed, that fits the height limit.

    @param
        **width_to_fit**
        The widht you want to fit, in normal pixels (1920 is the scene 
        width). These pixels will be processed to manim dimensions.
    """
    width_to_fit = ManimDimensions.width_to_manim_width(width_to_fit)

    image_width_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_width(width_to_fit)
    image_height_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_height(width_to_fit)

    # As it is a 16:9 proportion, the height is the measure that limits the most
    if image_height_fitted.width < image_width_fitted.width:
        return image_height_fitted
    
    return image_width_fitted

def fullscreen_image(filename, image_mode: str = 'RGBA', **kwargs):
    """
    Returns an ImageMobject that fits whole screen dimensions (1920x1080). It
    will ignore the dimension that is out of bounds.
    """
    image_width_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_width(ManimDimensions.width_to_manim_width(1920))

    # We want the image that occupies the whole screen
    if ManimDimensions.manim_height_to_height(image_width_fitted.height) >= 1080:
        return image_width_fitted
    
    image_height_fitted = ImageMobject(filename, image_mode, **kwargs).scale_to_fit_height(ManimDimensions.height_to_manim_height(1080))

    return image_height_fitted

def preprocess_image(image: ImageMobject):
    """
    This method processes images bigger than our 1920x1080 dimensions and returns it
    scaled down to fit those dimensions. You should use this method as the first one
    when working with ImageMobjects, and then scaling it down as much as you need.
    """
    if ManimDimensions.manim_width_to_width(image.width) > 1920:
        image.scale_to_fit_width(ManimDimensions.width_to_manim_width(1920))
    if ManimDimensions.manim_height_to_height(image.height) > 1080:
        image.scale_to_fit_height(ManimDimensions.height_to_manim_height(1080))

    return image


from yta_multimedia.video.generation.manim.classes.base_three_d_manim_animation import BaseThreeDManimAnimation
from manim import *


class Axes3DExample(BaseThreeDManimAnimation):
    def construct(self):
        axes = ThreeDAxes()

        x_label = axes.get_x_axis_label(Tex("x"))
        y_label = axes.get_y_axis_label(Tex("y")).shift(UP * 1.8)

        # 3D variant of the Dot() object
        dot = Dot3D()

        # zoom out so we see the axes
        self.set_camera_orientation(zoom=0.5)

        self.play(FadeIn(axes), FadeIn(dot), FadeIn(x_label), FadeIn(y_label))

        self.wait(0.5)

        # animate the move of the camera to properly see the axes
        self.move_camera(phi=75 * DEGREES, theta=30 * DEGREES, zoom=1, run_time=1.5)

        # built-in updater which begins camera rotation
        self.begin_ambient_camera_rotation(rate=0.15)

        # one dot for each direction
        upDot = dot.copy().set_color(RED)
        rightDot = dot.copy().set_color(BLUE)
        outDot = dot.copy().set_color(GREEN)

        self.wait(1)

        self.play(
            upDot.animate.shift(UP),
            rightDot.animate.shift(RIGHT),
            outDot.animate.shift(OUT),
        )

        self.wait(2)

class Image3D(BaseThreeDManimAnimation):
    def construct(self):
        self.video1 = ImageMobject(
                filename_or_array = 'C:/Users/dania/Desktop/wallpaper1080.png',
            ).scale_to_fit_height(3)
        ax = Axes(
            x_range=[0, 10, 1],
            x_length=9,
            y_range=[0, 20, 5],
            y_length=6,
            axis_config={"include_numbers": True, "include_tip": False},

        ).to_edge(DL + RIGHT + UP, buff=1).scale(0.7)
        labels = ax.get_axis_labels()

        self.play(Create(VGroup(ax, labels)))
        self.play(FadeIn(self.video1))
        self.wait(3)
        self.move_camera(phi=0*DEGREES, theta= -90 * DEGREES, zoom= 0.7, run_time=0.4, gamma=0*DEGREES)

        self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='phi')
        self.begin_ambient_camera_rotation(90 * DEGREES / 3, about='theta')
        self.wait(3)


from yta_multimedia.video.parser import VideoParser
from yta_general_utils.image.converter import ImageConverter
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from moviepy import VideoClip
from imageio import imsave
from typing import Union

import numpy as np


class FrameExtractionType(Enum):
    """
    Enum to simplify the way we indicate the frame extraction
    operation we want to execute.
    """
    TIME = 'time'
    FRAME_NUMBER = 'frame_number'

class VideoFrameExtractor:
    """
    Class to simplify the process of extracting video frames by time
    or frame number.

    A moviepy clip is built by consecutive frames. The first frame is
    on t = 0, and the next frames are obtained by applying a (a, b]
    interval. This means that if we a have a video of 10 fps, each 
    frame will last 0.1s. So, considering the previous condition, the
    second frame will be at 0.1s, so you will be able to access it by
    providing a time between (0.000000000001, 0.1].
    """
    # # TODO: Implement the 'output_filename' optional parameter to store
    # # the frames if parameter is provided.
    # @classmethod
    # def _get_frame_time(cls, video, frame_number_or_time, extraction_type: FrameExtractionType):
    #     """
    #     Returns the time to obtain the frame of the provided 'video'
    #     according to the provided 'frame_number_or_time' parameter and
    #     the desired 'extraction_type'.

    #     This method is useful to be used first to validate that the
    #     frame times or numbers are valid and avoid getting all the 
    #     frames and loading them in memory if it is going to fail.
    #     """
    #     video = VideoParser.to_moviepy(video)
    #     extraction_type = FrameExtractionType.to_enum(extraction_type)
        
    #     if not NumberValidator.is_positive_number(frame_number_or_time):
    #         raise Exception('The provided "frame_number_or_time" is not a valid number.')
        
    #     if extraction_type == FrameExtractionType.FRAME_NUMBER:
    #         max_frame_number = video.fps * video.duration
    #         if frame_number_or_time > max_frame_number:
    #             raise Exception(f'The provided "frame_number" parameter {str(frame_number_or_time)}" is not valid. The maximum is {str(max_frame_number)}.')
            
    #         frame_number_or_time = frame_number_or_time * 1.0 / video.fps
    #     elif extraction_type == FrameExtractionType.TIME:
    #         if frame_number_or_time > video.duration:
    #             raise Exception(f'The provided "time" parameter {str(frame_number_or_time)} is not valid. The maximum is {str(video.duration)}.')
            
    #     return frame_number_or_time

    # # TODO: I think I won't use this because I need to validate them first
    # @classmethod
    # def _get_frame(cls, video, frame_number_or_time, extraction_type: FrameExtractionType, output_filename: str = None):
    #     """
    #     Returns the frame of the provided 'video' according to the provided
    #     'frame_number_or_time' parameter and the desired 'extraction_type'.
    #     """
    #     video = VideoParser.to_moviepy(video)

    #     frame = video.get_frame(t = cls._get_frame_time(video, frame_number_or_time, extraction_type))

    #     if output_filename:
    #         cls._save_frame(frame, output_filename)

    #     return frame
    
    # @classmethod
    # def _get_frames(cls, video, frame_numbers_or_times, extraction_type: FrameExtractionType):
    #     """
    #     This method validates if all the provided 'frame_numbers_or_times'
    #     array parameter items are valid or not and them get all the frames
    #     and returns them as an array. This is optimum as it doesn't load
    #     frames until it's been verified the condition that all are valid.
    #     """
    #     video = VideoParser.to_moviepy(video)

    #     # Validate and get all times to obtain the frames later
    #     times = [cls._get_frame_time(video, frame_number_or_time, extraction_type) for frame_number_or_time in frame_numbers_or_times]

    #     return [video.get_frame(t = time) for time in times]
    
    # @classmethod
    # def get_frames_by_time(cls, video, times: list[float]):
    #     """
    #     Returns the frames (as numpy arrays) from the provided 'video'
    #     that corresponds to the provided 'times' time momnets of the
    #     video.

    #     This method will raise an Exception if the limits are not valid
    #     for the given 'video'.
    #     """
    #     return cls._get_frames(video, times, FrameExtractionType.TIME)

    # @classmethod
    # def _save_frame(cls, frame, output_filename: str):
    #     if not output_filename:
    #         return None
        
    #     # TODO: Validate 'output_filename'
        
    #     # TODO: This is how 'video.save_frame' works
    #     # https://github.com/Zulko/moviepy/blob/master/moviepy/video/VideoClip.py#L166
    #     # from imageio import imsave
    #     # im = self.get_frame(t)
    #     # if with_mask and self.mask is not None:
    #     #     mask = 255 * self.mask.get_frame(t)
    #     #     im = np.dstack([im, mask]).astype("uint8")
    #     # else:
    #     #     im = im.astype("uint8")

    #     imsave(output_filename, frame.astype("uint8"))

    #     return True

    # @classmethod
    # def get_frame_by_time(cls, video, time: float, output_filename: str = None):
    #     """
    #     Returns the frame (as numpy arrays) from the provided 'video'
    #     that corresponds to the provided 'time' moment.

    #     This method will raise an Exception if the limits are not valid
    #     for the given 'video'.
    #     """
    #     frame = cls.get_frames_by_time(video, [time])[0]

    #     if output_filename:
    #         cls._save_frame(frame, output_filename)

    #     return frame
    
    # @classmethod
    # def get_frames_by_time_from_start_to_end(cls, video, start_time: float, end_time: float):
    #     """
    #     Returns the frames (as numpy arrays) from the provided 'video'
    #     that corresponds to the lapse of time between the provided
    #     'start_time' and 'end_time'.

    #     This method will raise an Exception if the limits are not valid
    #     for the given 'video'.
    #     """
    #     video = VideoParser.to_moviepy(video)

    #     # Validate start and end parameters
    #     if not NumberValidator.is_positive_number(start_time) or start_time < 0 or start_time > video.duration:
    #         raise Exception(f'The provided "start_time" parameter "{str(start_time)}" is not valid. It must be a positive number between 0 and the video duration ({str(video.duration)}).')
    #     if not NumberValidator.is_positive_number(end_time) or end_time < 0 or end_time < start_time or end_time > video.duration:
    #         raise Exception(f'The provided "end_time" parameter "{str(end_time)}" is not valid. It must be a positive number between 0 and the video duration ({str(video.duration)}), and greater than the provided "start_time" parameter.')
        
    #     total_frames = int(video.duration * video.fps)
    #     time_moments = []
    #     for frame in range(total_frames):
    #         frame_time_moment = frame / video.fps  # tiempo en segundos para el frame actual
    #         if start_time <= frame_time_moment <= end_time:
    #             time_moments.append(frame_time_moment)

    #     return cls.get_frames_by_time(video, time_moments)

    # @classmethod
    # def get_frame_by_frame_number(cls, video, frame_number: int, output_filename: str = None):
    #     """
    #     Returns the frame (as numpy array) for the provided 'video' that
    #     corresponds to the provided 'frame_number'.

    #     This method will raise an Exception if the provided 'frame_number'
    #     is not a valid frame number for the given 'video'.
    #     """
    #     frame = cls._get_frame(video, frame_number, FrameExtractionType.FRAME_NUMBER)

    #     if output_filename:
    #         cls._save_frame(frame, output_filename)

    #     return frame
    
    # @classmethod
    # def get_frames_by_frame_number(cls, video, frame_numbers: list[int]):
    #     """
    #     Returns the frames (as numpy arrays) from the provided 'video'
    #     that corresponds to the provided 'frame_numbers'.

    #     This method will raise an Exception if the limits are not valid
    #     for the given 'video'.
    #     """
    #     return cls._get_frames(video, frame_numbers, FrameExtractionType.FRAME_NUMBER)
    
    # @classmethod
    # def get_frames_by_frame_number_from_start_to_end(cls, video, start_frame: float, end_frame: float):
    #     """
    #     Returns the frames (as numpy arrays) from the provided 'video' 
    #     from the 'start_frame' to the 'end_frame'.

    #     This method will raise an Exception if the limits are not valid
    #     for the given 'video'.
    #     """
    #     video = VideoParser.to_moviepy(video)

    #     max_frame_number = int(video.fps * video.duration)
    #     # Validate start and end parameters
    #     if not NumberValidator.is_positive_number(start_frame) or start_frame < 0 or start_frame > max_frame_number:
    #         raise Exception(f'The provided "start_frame" parameter "{str(start_frame)}" is not valid. It must be a positive number between 0 and the maximum frame ({str(max_frame_number)}).')
    #     if not NumberValidator.is_positive_number(end_frame) or end_frame < 0 or end_frame < start_frame or end_frame > max_frame_number:
    #         raise Exception(f'The provided "end_frame" parameter "{str(end_frame)}" is not valid. It must be a positive number between 0 and the maximum frame ({str(max_frame_number)}), and greater than the provided "start_time" parameter.')
        
    #     frames = [frame for frame in range(start_frame, end_frame + 1)]
    #     time_moments = [cls._get_frame_time(video, frame, FrameExtractionType.FRAME_NUMBER) for frame in frames]

    #     return cls.get_frames_by_time(video, time_moments)

    @staticmethod
    def get_frame_by_number(video: VideoClip, t: int = 0, output_filename: Union[str, None] = None):
        """
        Get the frame 't' of the provided 'video'. The frame number
        must be a valid one.
        """
        frame = VideoFrameExtractor.get_frames_by_number(video, [t])

        if output_filename:
            # TODO: Why type (?)
            #imsave(output_filename, frame.astype("uint8"))
            imsave(output_filename, frame)

        return frame

    @staticmethod
    def get_frames_by_number(video, t: list[int] = [0]):
        """
        Get all the 't' frames of the provide d'video'. Those frame
        numbers must be valid.
        """
        if not PythonValidator.is_list(t):
            if NumberValidator.is_positive_number(t):
                t = [t]
            else:
                raise Exception('The provided "t" is not an array of frame numbers nor a single one.')
            
        return [VideoFrameExtractor.get_frame(video, FrameExtractionType.FRAME_NUMBER, t_) for t_ in t]
    
    @staticmethod
    def get_frame_by_time(video: VideoClip, t: float = 0.0, output_filename: Union[str, None] = None):
        """
        Get the frame in the provided 't' time moment of the
        provided 'video'. The frame time must be a valid one,
        between 0 and the video duration.
        """
        frame = VideoFrameExtractor.get_frames_by_time(video, [t])[0]

        if output_filename:
            # TODO: Why type (?)
            #imsave(output_filename, frame.astype("uint8"))
            imsave(output_filename, frame)

        return frame

    @staticmethod
    def get_frames_by_time(video, t: list[float] = [0.0]):
        """
        Get all the frame corresponding to the time moments in
        provided 't' of the provided 'video'. All frame times
        must be valid, between 0 and the video duration.
        """
        if not PythonValidator.is_list(t):
            if NumberValidator.is_positive_number(t):
                t = [t]
            else:
                raise Exception('The provided "t" is not an array of frame times nor a single one.')
            
        return [VideoFrameExtractor.get_frame(video, FrameExtractionType.TIME, t_) for t_ in t]
    
    @staticmethod
    def get_frame(video, mode: FrameExtractionType = FrameExtractionType.FRAME_NUMBER, t: Union[float, int] = 0):
        """
        Get the frame corresponding to the provided 't' frame number
        or time moment. Feel free to use the moviepy .get_frame()
        method instead.

        This is my own method due to some problems with the original
        moviepy .get_frame() and because of its laxity.
        """
        video = VideoParser.to_moviepy(video)
        mode = FrameExtractionType.to_enum(mode)

        if not NumberValidator.is_positive_number(t):
            raise Exception('The provided "t" value is not a positive number.')

        if mode == FrameExtractionType.FRAME_NUMBER:
            t = int(t)
        elif mode == FrameExtractionType.TIME:
            t = float(t)

        frames = VideoFrameExtractor.get_all_frames(video)
        duration = video.duration
        fps = video.fps

        if mode == FrameExtractionType.FRAME_NUMBER and t > len(frames):
            raise Exception(f'The provided "t" frame number "{str(t)}" is not valid.. The maximum is {str(len(frames - 1))}.')
        elif mode == FrameExtractionType.TIME and t > duration:
            raise Exception(f'The provided "t" frame duration "{str(t)}" is not valid. The maximum is {str(duration)}.')

        if mode == FrameExtractionType.FRAME_NUMBER:
            frame = frames[t]
        elif mode == FrameExtractionType.TIME:
            if t == duration:
                frame = frames[-1]
            else:
                frames_number = int(fps * duration)
                frame_duration = duration / frames_number
                # We add a 10% of frame_duration to make sure it fits
                # the expected frame
                frame = frames[int(t * fps + frame_duration * 0.1)]

        return frame

    @classmethod
    def get_all_frames(cls, video):
        """
        Returns all the frames from the provided 'video'.
        """
        video = VideoParser.to_moviepy(video)

        return [frame for frame in video.iter_frames()]
    
    @staticmethod
    def get_first_frame(video):
        """
        Obtain the first frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        return VideoFrameExtractor.get_all_frames(video)[0]
    
    @staticmethod
    def get_last_frame(video):
        """
        Obtain the last frame of the provided 'video' as a ndim=3
        numpy array containing the clip part (no mask) as not
        normalized values (between 0 and 255).
        """
        video = VideoParser.to_moviepy(video)

        return VideoFrameExtractor.get_all_frames(video)[-1]
    
    # TODO: Would be perfect to have some methods to turn frames into
    # RGBA denormalized (0, 255) or normalized (0, 1) easier because
    # it is needed to work with images and other libraries. Those 
    # methods would iterate over the values and notice if they are in
    # an specific range so they need to be change or even if they are
    # invalid values (not even in [0, 255] range because they are not
    # rgb or rgba colors but math calculations).
    # This is actually being done by the VideoMaskHandler
    @classmethod
    def get_frame_as_rgba_by_time(cls, video, time: int, do_normalize: bool = False, output_filename: str = None):
        """
        Gets the frame of the requested 'time' of the provided
        'video' as a normalized RGBA numpy array that is built
        by joining the rgb frame (from main clip) and the alpha
        (from .mask clip), useful to detect transparent regions.
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)

        # We first normalize the clips
        main_frame = cls.get_frame_by_time(video, time) / 255  # RGB numpy array normalized 3d <= r,g,b
        mask_frame = cls.get_frame_by_time(video.mask, time)[:, :, np.newaxis]  # Alpha numpy array normalized 1d <= alpha
        # Combine RGB of frame and A from mask to RGBA numpy array (it is normalized)
        frame_rgba = np.concatenate((main_frame, mask_frame), axis = 2) # 4d <= r,g,b,alpha

        if output_filename:
            # TODO: Check extension
            ImageConverter.numpy_image_to_pil(frame_rgba).save(output_filename)
            # TODO: Write numpy as file image
            # Video mask is written as 0 or 1 (1 is transparent)
            # but main frame is written as 0 to 255, and the
            # 'numpy_image_to_pil' is expecting from 0 to 1
            # (normalized) instead of from 0 to 255 so it won't
            # work

        if not do_normalize:
            frame_rgba *= 255

        return frame_rgba
    


"""
    In this file we register all video greenscreens we have
    available in our system. These greenscreens have their
    own resource uploaded to Google Drive and we have also
    detected the greenscreen color and the pixels in which
    the area is and we define those parameters below.

    This file is for letting these greenscreens to be used
    by the VideoGreenscreen class to inject some pictures
    or videos on it and enhance the video experience.

    TODO: Read below
    This should be replaced by a database in which we store
    this information and keep it updated, not directly in
    code, but for now it is our system.
"""
from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.classes.greenscreen_area_details import GreenscreenAreaDetails
from yta_multimedia.greenscreen.enums import GreenscreenType


LAPTOP_ON_TABLE_VIDEO_GREENSCREEN = GreenscreenDetails(
    greenscreen_areas = [
        GreenscreenAreaDetails(
            rgb_color = (0, 174, 62),
            similar_greens = [],
            upper_left_pixel = (569, 229),
            lower_right_pixel = (1385, 737),
            frames = None
        )
    ],
    filename_or_google_drive_url = 'https://drive.google.com/file/d/1_luaJcNqP49AzeXuF4LPpAq7LvNzr95O/view?usp=sharing',
    type = GreenscreenType.VIDEO
)


from yta_general_utils.file.checker import FileValidator
from yta_general_utils.file.enums import FileType
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import VideoFileClip
from moviepy.Clip import Clip
from typing import Union


class VideoParser:
    """
    Class to simplify the way we parse video parameters.
    """
    @classmethod
    def to_moviepy(cls, video: Union[str, Clip], do_include_mask: bool = False):
        """
        This method is a helper to turn the provided 'video' to a moviepy
        video type. If it is any of the moviepy video types specified in
        method declaration, it will be returned like that. If not, it will
        be load as a VideoFileClip if possible, or will raise an Exception
        if not.

        The 'do_include_mask' parameter includes the mask in the video if
        True value provided. The 'do_check_duration' parameter checks and
        updates the real video duration to fix a bug in moviepy lib.
        """
        if not video:
            raise Exception('No "video" provided.')
        
        # TODO: Maybe check if subclass of VideoClip
        if not PythonValidator.is_string(video) and not PythonValidator.is_instance(video, [Clip]):
            raise Exception('The "video" parameter provided is not a valid type. Check valid types in method declaration.')
        
        if PythonValidator.is_string(video):
            if not filename_is_type(video, FileType.VIDEO):
                raise Exception('The "video" parameter provided is not a valid video filename.')
            
            if not FileValidator.file_is_video_file(video):
                raise Exception('The "video" parameter is not a valid video file.')
            
            video = VideoFileClip(video, has_mask = do_include_mask)

        # TODO: This below just adds a mask attribute but
        # without fps and empty, so it doesn't make sense
        # if do_include_mask and not video.mask:
        #     video = video.add_mask()

        return video